from __future__ import annotations

import tempfile
import warnings
from contextlib import contextmanager
from os import PathLike
from typing import Hashable, Iterator, Sequence

import cv2
import imageio
import numpy as np
import soundfile as sf
from diskcache import Cache
from imageio.core.format import Format
from tqdm import tqdm

from ..attribute import Attribute, AttributeType
from ..effect import Effect
from ..enum import BlendingMode, CacheType, Direction
from ..imgproc import alpha_composite
from ..transform import Transform, TransformValue
from .protocol import AUDIO_BLOCK_SIZE, AUDIO_SAMPLING_RATE, AudioLayer, Layer


class Composition:
    """A base layer that integrates multiple layers into one video.

    Users create a composition by specifying both time and resolution. Next, multiple layers can be added to
    the target composition through ``Composition.add_layer()``. During this process, additional information such as
    the layer's name, start time, position, opacity, and drawing mode can be specified.
    Finally, the composition integrates the layers in the order they were added to create a single video.

    Another composition can also be added as a layer within a composition.
    By nesting compositions in this way, more complex motions can be created.

    Examples:
        >>> import movis as mv
        >>> composition = mv.layer.Composition(size=(640, 480), duration=5.0)
        >>> composition.add_layer(
        ...     mv.layer.Rectangle(size=(640, 480), color=(127, 127, 127), duration=5.0),
        ...     name='bg')
        >>> len(composition)
        1
        >>> composition['bg'].opacity.enable_motion().extend(
        ...     keyframes=[0, 1, 2, 3, 4],
        ...     values=[1.0, 0.0, 1.0, 0.0, 1.0],
        ...     easings=['ease_out5'] * 5)
        >>> composition.write_video('output.mp4')

    Args:
        size:
            A tuple representing the size of the composition in the form of ``(width, height)``.
        duration:
            The duration along the time axis for the composition.
    """

    def __init__(
        self, size: tuple[int, int] = (1920, 1080), duration: float = 1.0
    ) -> None:
        self._layers: list[LayerItem] = []
        self._name_to_layer: dict[str, LayerItem] = {}
        assert duration > 0, "duration must be positive"
        self._duration = duration
        self._cache: Cache = Cache(size_limit=1024 * 1024 * 1024)
        self._preview_level: int = 1
        assert isinstance(size, tuple) and len(size) == 2, "size must be a tuple of length 2"
        assert size[0] > 0 and size[1] > 0, "size must be positive"
        self._size = size

    @property
    def size(self) -> tuple[int, int]:
        """The size of the composition in the form of ``(width, height)``."""
        return self._size

    @property
    def duration(self) -> float:
        """The duration of the composition."""
        return self._duration

    @property
    def preview_level(self) -> int:
        """The resolution of the rendering of the composition.
        For example, if ``preview_level=2`` is set,
        the composition's resolution is ``(W / 2, H / 2)``."""
        return self._preview_level

    @preview_level.setter
    def preview_level(self, level: int) -> None:
        assert level > 0, "preview_level must be greater than 0"
        self._preview_level = int(level)
        if self._preview_level != level:
            self._cache.clear()

    @contextmanager
    def preview(self, level: int = 2) -> Iterator[None]:
        """Context manager method to temporarily change the ``preview_level`` using the `with` syntax.

        For example, ``with self.preview(level=2):`` would change the ``preview_level`` to 2 in that scope.

        Args:
            level:
                The resolution of the rendering of the composition.
                For example, if ``level=2`` is set, the composition's resolution is ``(W / 2, H / 2)``.

        Examples:
            >>> import movis as mv
            >>> composition = mv.layer.Composition(size=(640, 480), duration=5.0)
            >>> with composition.preview(level=2):
            ...     image = composition(0.0)
            >>> image.shape
            (240, 320, 4)
            >>> with composition.preview(level=4):
            ...     image = composition(0.0)
            >>> image.shape
            (120, 160, 4)
        """
        assert level > 0
        original_level = self.preview_level
        self.preview_level = level
        try:
            yield
        finally:
            self.preview_level = original_level

    @property
    def layers(self) -> Sequence[LayerItem]:
        """Returns a list of ``LayerItem`` objects."""
        return self._layers

    def keys(self) -> list[str]:
        """Returns a list of layer names.

        .. note::
            The keys are sorted in the order in which they will be rendered.

        Returns:
            A list of layer names sorted in the rendering order.
        """
        return [layer.name for layer in self._layers]

    def values(self) -> list[LayerItem]:
        """Returns a list of LayerItem objects.

        .. note::
            The elements of the list are not the layers themselves,
            but ``LayerItem`` containing information of the layers.

        Returns:
            A list of ``LayerItem`` objects.
        """
        return self._layers

    def items(self) -> list[tuple[str, LayerItem]]:
        """Returns a list of tuples, each consisting of a layer name and its corresponding item.

        Returns:
            A list of tuples, where each tuple contains a layer name and its layer item.
        """
        return [(layer.name, layer) for layer in self._layers]

    def __len__(self) -> int:
        return len(self._layers)

    def __getitem__(self, key: str) -> LayerItem:
        return self._name_to_layer[key]

    def __contains__(self, key: str) -> bool:
        return key in self._name_to_layer

    def __setitem__(self, key: str, value: LayerItem | Layer) -> None:
        if isinstance(value, LayerItem):
            self._layers.append(value)
            self._name_to_layer[key] = value
        elif callable(value):
            self.add_layer(value, name=key)
        else:
            raise ValueError("value must be LayerItem or Layer (i.e., callable)")

    def __delitem__(self, key: str) -> None:
        self.pop_layer(key)

    def get_key(self, time: float) -> tuple[Hashable, ...] | None:
        """Returns a tuple of hashable keys representing the state for each layer at the given time."""
        if time < 0.0 or self.duration <= time:
            return None
        layer_keys: list[Hashable] = [CacheType.COMPOSITION]
        for layer_item in self._layers:
            layer_time = time - layer_item.offset
            if layer_time < layer_item.start_time or layer_item.end_time <= layer_time:
                layer_keys.append(None)
            else:
                layer_keys.append(layer_item.get_key(time))
        return tuple(layer_keys)

    def __repr__(self) -> str:
        return f"Composition(size={self.size}, duration={self.duration}, layers={self._layers!r})"

    def add_layer(
        self,
        layer: Layer,
        name: str | None = None,
        position: tuple[float, float] | None = None,
        scale: float | tuple[float, float] | np.ndarray = (1.0, 1.0),
        rotation: float = 0.0,
        opacity: float = 1.0,
        blending_mode: BlendingMode | str = BlendingMode.NORMAL,
        anchor_point: float | tuple[float, float] | np.ndarray = (0.0, 0.0),
        origin_point: Direction | str = Direction.CENTER,
        transform: Transform | None = None,
        offset: float = 0.0,
        start_time: float = 0.0,
        end_time: float | None = None,
        audio_level: float = 0.0,
        visible: bool = True,
        audio: bool = True,
    ) -> LayerItem:
        """Add a layer to the composition.

        This method appends the target layer to the composition, along with details about the layer such as
        position, scale, opacity, and rendering mode. The composition registers layers wrapped within
        a ``LayerItem`` object, which consolidates these related details.
        Users can also add the layer with a unique name to the composition.
        In this case, the LayerItem can be accessed using ``composition['layer_name']``.
        To access the layer directly, users can reference it as ``composition['layer_name'].layer``,
        or ideally, retain it in a separate variable before registering with ``add_layer()``.

        A composition can also be treated as a layer.
        This means that users can embed one composition within another using the ``add_layer()`` method.
        This allows for more intricate image compositions and animations.

        Args:
            layer:
                An instance or function of the layer to be added to the composition,
                conforming to the ``Layer`` protocol.
            name:
                The unique name for the layer within the composition.
            position:
                The position of the layer. If unspecified,
                the layer is placed at the center of the composition by default.
            scale:
                Scale ``(sx, sy)`` of the layer. Defaults to ``(1.0, 1.0)``.
            rotation:
                Clockwise rotation angle (in degrees) of the layer. Default is ``0.0``.
            opacity:
                Opacity of the layer. Default is ``1.0``.
            blending_mode:
                Rendering mode of the layer. Can be specified as an Enum from ``BlendingMode``
                or as a string. Defaults to ``BlendingMode.NORMAL``.
            anchor_point:
                Defines the origin of the layer's coordinate system.
                The origin is determined by the sum of ``origin_point`` and ``anchor_point``.
                If ``origin_point`` is ``Direction.CENTER`` and ``anchor_point`` is ``(0, 0)``,
                the origin is the center of the layer. Default is ``(0, 0)``.
            origin_point:
                Initial reference point for the layer's coordinate system.
                The final origin is determined by the sum of ``origin_point`` and ``anchor_point``.
                Defaults to ``Direction.CENTER`` (center of the layer).
            transform:
                A Transform object managing the geometric properties and rendering mode of the layer.
                If specified, the arguments for ``position``, ``scale``, ``rotation``, ``anchor_point``,
                ``origin_point``, and ``blending_mode`` in ``add_layer()`` are ignored
                in favor of the values in ``transform``.
            offset:
                The starting time of the layer. For example, if ``start_time=0.0`` and ``offset=1.0``,
                the layer will appear after 1 second in the composition.
            start_time:
                The start time of the layer. This variable is used to clip the layer in the time axis direction.
                For example, if ``start_time=1.0`` and ``offset=0.0``, this layer will appear immediately
                with one second skipped.
            end_time:
                The end time of the layer. This variable is used to clip the layer in the time axis direction.
                For example, if ``start_time=0.0``, ``end_time=1.0``, and ``offset=0.0``,
                this layer will disappear after one second. If not specified,
                the layer's duration is used for ``end_time``.
            audio_level:
                The audio level of the layer (dB). If the layer has no audio, this value is ignored.
            visible:
                A flag specifying whether the layer is visible or not;
                if ``visible=False``, the layer in the composition is not rendered.
            audio:
                A flag specifying whether the audio is enabled or not;
                if ``audio=False``, the audio of the given layer is not used.

        Returns:
            A ``LayerItem`` object that wraps the layer and its corresponding information.
        """
        if name is None:
            name = f"layer_{len(self._layers)}"
        if name in self._name_to_layer:
            raise KeyError(f"Layer with name {name} already exists")
        end_time = end_time if end_time is not None else getattr(layer, "duration", 1e6)

        if position is None:
            position = self.size[0] / 2, self.size[1] / 2
        if transform is None:
            transform = Transform(
                position=position,
                scale=scale,
                rotation=rotation,
                opacity=opacity,
                anchor_point=anchor_point,
                origin_point=origin_point,
                blending_mode=blending_mode,
            )
        layer_item = LayerItem(
            layer,
            name=name,
            transform=transform,
            offset=offset,
            start_time=start_time,
            end_time=end_time,
            audio_level=audio_level,
            visible=visible,
            audio=audio,
        )
        self._layers.append(layer_item)
        self._name_to_layer[name] = layer_item
        self._cache.clear()
        return layer_item

    def pop_layer(self, name: str) -> LayerItem:
        """Removes a layer item from the composition and returns it.

        Args:
            name: The name of the layer to be removed.

        Returns:
            The layer item that was removed.
        """
        if name not in self._name_to_layer:
            raise KeyError(f"Layer with name {name} does not exist")
        index = next(i for i in range(len(self._layers)) if self._layers[i].name == name)
        layer_item = self._layers.pop(index)
        self._name_to_layer.pop(name)
        self._cache.clear()
        return layer_item

    def clear(self) -> None:
        """Removes all layers from the composition."""
        self._layers.clear()
        self._name_to_layer.clear()
        self._cache.clear()

    def __call__(self, time: float) -> np.ndarray | None:
        if time < 0.0 or self.duration <= time:
            return None

        L = self._preview_level
        current_shape = self.size[1] // L, self.size[0] // L

        key = self.get_key(time)
        if key in self._cache:
            cached_frame: np.ndarray = self._cache[key]
            if cached_frame.shape[:2] == current_shape:
                return cached_frame
            else:
                del self._cache[key]

        frame = np.zeros(current_shape + (4,), dtype=np.uint8)
        for layer_item in self._layers:
            frame = layer_item._composite(
                frame, time, preview_level=self._preview_level,
                cache=self._cache)
        self._cache[key] = frame
        return frame

    def get_audio(self, start_time: float, end_time: float) -> np.ndarray | None:
        """Returns the audio of the composition as a numpy array.

        Args:
            start_time:
                The start time of the audio. This variable is used to clip the audio in the time axis direction.
            end_time:
                The end time of the audio. This variable is used to clip the audio in the time axis direction.
                If not specified, the composition's duration is used for ``end_time``.

        Returns:
            The audio of the composition as a numpy array. If no audio is found, ``None`` is returned.
        """
        assert start_time >= 0, "start_time must be nonnegative"
        assert end_time is None or start_time < end_time
        target_layers = [li for li in self.layers if hasattr(li.layer, 'get_audio')]

        audio = None
        for layer_item in target_layers:
            result = layer_item._get_audio_data(start_time, end_time)
            if result is None:
                continue
            layer_time_start, _, audio_i = result
            time_i = layer_time_start + layer_item.offset - start_time
            ind_start = int(time_i * AUDIO_SAMPLING_RATE)
            if audio is None:
                n_samples = int((end_time - start_time) * AUDIO_SAMPLING_RATE)
                audio = np.zeros((2, n_samples), dtype=np.float32)
            ind_end = min(ind_start + audio_i.shape[1], audio.shape[1])
            length = ind_end - ind_start
            audio[:, ind_start:ind_end] += audio_i[:, :length]
        return audio

    def _compute_frame(self, time: float) -> np.ndarray:
        return np.asarray(self(time))

    def _write_video(
        self, start_time: float, end_time: float,
        fps: float, writer: Format.Writer,
    ) -> None:
        times = np.arange(start_time, end_time, 1.0 / fps)
        for t in tqdm(times, total=len(times)):
            frame = np.asarray(self(t))
            writer.append_data(frame)
        writer.close()

    def write_video(
        self,
        dst_file: str | PathLike,
        start_time: float = 0.0,
        end_time: float | None = None,
        codec: str = "libx264",
        pixelformat: str = "yuv420p",
        input_params: list[str] | None = None,
        output_params: list[str] | None = None,
        fps: float = 30.0,
        audio: bool = True,
        audio_codec: str | None = None,
    ) -> None:
        """Writes the composition's contents to a video file.

        Args:
            dst_file:
                The path to the destination video file.
            start_time:
                The start time of the video. This variable is used to clip the video in the time axis direction.
            end_time:
                The end time of the video. This variable is used to clip the video in the time axis direction.
                If not specified, the composition's duration is used for ``end_time``.
            codec:
                The codec used to encode the video. Default is ``libx264``.
            pixelformat:
                The pixel format of the video. Default is ``yuv420p``.
            input_params:
                A list of parameters to be passed to the ffmpeg input.
            output_params:
                A list of parameters to be passed to the ffmpeg output.
                For example, ``["-preset", "medium"]`` or ``["-crf", "18"]``.
            fps:
                The frame rate of the video. Default is ``30.0``.
            audio:
                A flag specifying whether to include audio in the video.
                If ``audio=False``, the audio of the composition is not rendered.
                Note that if the composition has no audio, this value is simply ignored and
                the video is rendered without audio.
            audio_codec:
                The codec used to encode the audio. If not specified, the default codec
                determined by ``codec`` is used. For example, if ``codec="libx264"``,
                the default value of ``audio_codec`` is ``aac``.
        """
        if end_time is None:
            end_time = self.duration
        writer = None
        if audio:
            with tempfile.NamedTemporaryFile(suffix='.wav') as audio_fp:
                audio_array = self.get_audio(start_time, end_time)
                if audio_array is None:
                    audio_path = None
                else:
                    sf.write(
                        audio_fp, audio_array.transpose(),
                        samplerate=AUDIO_SAMPLING_RATE,
                        subtype='PCM_16')
                    audio_path = audio_fp.name
                with warnings.catch_warnings():
                    # XXX: Suppress the deprecation warning from imageio-ffmpeg.
                    warnings.simplefilter("ignore", category=DeprecationWarning)
                    writer = imageio.get_writer(
                        uri=str(dst_file), fps=fps, codec=codec, pixelformat=pixelformat,
                        macro_block_size=None, ffmpeg_log_level="error",
                        input_params=input_params, output_params=output_params,
                        audio_path=audio_path, audio_codec=audio_codec)
                    self._write_video(start_time, end_time, fps, writer)
        else:
            with warnings.catch_warnings():
                # XXX: Suppress the deprecation warning from imageio-ffmpeg.
                warnings.simplefilter("ignore", category=DeprecationWarning)
                writer = imageio.get_writer(
                    uri=str(dst_file), fps=fps, codec=codec, pixelformat=pixelformat,
                    input_params=input_params, output_params=output_params,
                    macro_block_size=None, ffmpeg_log_level="error")
                self._write_video(start_time, end_time, fps, writer)

    def render_and_play(
        self,
        start_time: float = 0.0,
        end_time: float | None = None,
        fps: float = 30.0,
        preview_level: int = 2
    ) -> None:
        """Renders the composition and plays it in a Jupyter notebook.

        .. note::
            This method requires ``ipywidgets`` and ``IPython`` to be installed.
            These are usually installed automatically when users install Jupyter notebook.

        Args:
            start_time:
                The start time of the video. This variable is used to clip the video in the time axis direction.
            end_time:
                The end time of the video. This variable is used to clip the video in the time axis direction.
            fps:
                The frame rate of the video. Default is ``30.0``.
            preview_level:
                The resolution of the rendering of the composition.
                For example, if ``preview_level=2`` is set, the resolution of the output is ``(W / 2, H / 2)``.
                Default is ``2``.
        """
        from IPython.display import display
        from ipywidgets import Video

        if end_time is None:
            end_time = self.duration

        times = np.arange(start_time, end_time, 1.0 / fps)
        with tempfile.NamedTemporaryFile(suffix='.mp4') as fp:
            with self.preview(level=preview_level):
                filename: str = fp.name
                writer = imageio.get_writer(
                    filename, fps=fps, codec="libx264",
                    ffmpeg_params=["-preset", "veryfast"],
                    pixelformat="yuv444p", macro_block_size=None,
                    ffmpeg_log_level="error")
                for t in tqdm(times, total=len(times)):
                    frame = np.asarray(self(t))
                    writer.append_data(frame)
                writer.close()
                display(Video.from_file(filename, autoplay=True, loop=True))

    def write_audio(
        self,
        dst_file: str | PathLike,
        start_time: float = 0.0,
        end_time: float | None = None,
        format: str | None = None,
        subtype: str | None = None,
    ) -> None:
        """Writes the audio of the composition to a file.

        The current supported audio formats are limited to those supported by the soundfile module
        (wav, flac, ogg, etc.). For more details, please refer to the soundfile documentation.

        Args:
            dst_file:
                The path to the destination audio file.
            start_time:
                The start time of the audio. This variable is used to clip the audio in the time axis direction.
            end_time:
                The end time of the audio. This variable is used to clip the audio in the time axis direction.
                If not specified, the composition's duration is used for ``end_time``.
            format:
                The format of the audio file. If not specified, the format is inferred from the file extension.
            subtype:
                The subtype of the audio file. If not specified, the subtype is inferred from the file extension.
        """
        if end_time is None:
            end_time = self.duration
        audio_array = self.get_audio(start_time, end_time)
        if audio_array is None:
            raise ValueError("No audio found in the composition")
        sf.write(
            dst_file, audio_array.transpose(),
            samplerate=AUDIO_SAMPLING_RATE,
            format=format,
            subtype=subtype)


class LayerItem:
    """A wrapper layer for managing additional info. (e.g., the name and position) of each layer in a composition.

    Usually, there is no need for the user to create this layer directly.
    However, editing additional information like the layer's position or opacity,
    or when adding animations or effects, requires editing this layer.

    ``LayerItem`` can be accessed as the return value of ``composition.add_layer()``
    or by specifying it like ``composition['layer_name']``.
    If you want to directly access the layer, refer to the ``layer_item.layer`` property.

    Args:
        layer:
            The layer to be wrapped.
        name:
            The name of the layer. The layer name must be unique within the composition.
        transform:
            An instance of ``Transform`` that includes multiple properties
            used to transform the layer within the composition.
        offset:
            The starting time of the layer. For example, if ``start_time=0.0`` and ``offset=1.0``,
            the layer will appear after 1 second in the composition.
        start_time:
            The start time of the layer. This variable is used to clip the layer in the time axis direction.
            For example, if ``start_time=1.0`` and ``offset=0.0``, this layer will appear immediately
            with one second skipped.
        end_time:
            The end time of the layer. This variable is used to clip the layer in the time axis direction.
            For example, if ``start_time=0.0``, ``end_time=1.0``, and ``offset=0.0``,
            this layer will disappear after one second.
            If not specified, the layer's duration is used for ``end_time``.
        audio_level:
            The relative audio level of the layer (dB). If the layer has no audio, this value is ignored.
        visible:
            A flag specifying whether the layer is visible or not;
            if ``visible=False``, the layer in the composition is not rendered.
        audio:
            A flag specifying whether the audio is enabled or not;
            if ``audio=False``, the audio of the given layer is not used.
    """
    def __init__(
            self, layer: Layer, name: str = 'layer', transform: Transform | None = None,
            offset: float = 0.0, start_time: float = 0.0, end_time: float | None = None,
            audio_level: float = 0.0, visible: bool = True, audio: bool = True):
        self.layer: Layer = layer
        self.name: str = name
        self.transform: Transform = transform if transform is not None else Transform()
        self.offset: float = offset
        self.start_time: float = start_time
        self.end_time: float = end_time if end_time is not None else getattr(layer, "duration", 1e6)
        self.audio_level: Attribute = Attribute(audio_level, AttributeType.SCALAR, range=(-1000, 1000))
        self.visible: bool = visible
        self.audio: bool = audio
        self._effects: list[Effect] = []

    @property
    def duration(self) -> float:
        """The duration of the layer item.

        .. note::
            This value is determined by the difference between ``end_time`` and ``start_time``,
            not by the duration of the layer.
        """
        return self.end_time - self.start_time

    def add_effect(self, effect: Effect) -> Effect:
        """Adds an effect to the layer.

        Args:
            effect:
                The effect to be added to the layer.

        Returns:
            The effect that was added.
        """
        self._effects.append(effect)
        return effect

    def remove_effect(self, effect: Effect) -> None:
        """Removes an effect from the layer.

        Args:
            effect:
                The effect to be removed from the layer.
        """
        self._effects.remove(effect)

    @property
    def effects(self) -> list[Effect]:
        """A list of effects applied to the layer."""
        return self._effects

    @property
    def anchor_point(self) -> Attribute:
        """The anchor point of the layer."""
        return self.transform.anchor_point

    @property
    def position(self) -> Attribute:
        """The position of the layer."""
        return self.transform.position

    @property
    def scale(self) -> Attribute:
        """The scale of the layer."""
        return self.transform.scale

    @property
    def rotation(self) -> Attribute:
        """The rotation of the layer."""
        return self.transform.rotation

    @property
    def opacity(self) -> Attribute:
        """The opacity of the layer."""
        return self.transform.opacity

    @property
    def origin_point(self) -> Direction:
        """The origin point of the layer."""
        return self.transform.origin_point

    @property
    def blending_mode(self) -> BlendingMode:
        """The blending mode of the layer."""
        return self.transform.blending_mode

    def get_key(self, time: float) -> tuple[Hashable, Hashable, Hashable]:
        """Returns the state of the layer item at the given time.

        Args:
            time:
                The time at which the layer is rendered.

        Returns:
            A tuple of hashable keys representing the state of the layer at the given time."""
        if not self.visible:
            return (None, None, None)
        layer_time = time - self.offset
        transform_key = self.transform.get_current_value(layer_time)
        layer_key = self.layer.get_key(layer_time) if hasattr(self.layer, 'get_key') else layer_time

        def get_effect_key(e: Effect) -> Hashable | None:
            return e.get_key(layer_time) if hasattr(e, 'get_key') else layer_time

        effects_key = None if len(self._effects) == 0 else tuple([get_effect_key(e) for e in self._effects])
        return (transform_key, layer_key, effects_key)

    def _get_fg_image(self, time: float, cache: Cache | None = None) -> np.ndarray | None:
        key = None
        if cache is not None:
            key_body = self.get_key(time)
            key = (CacheType.LAYER, self.name, key_body)
            if key in cache:
                return cache[key]
        fg_image = self(time)
        if fg_image is None:
            return None
        assert isinstance(fg_image, np.ndarray), "Rendered layer image must be a numpy array"
        assert fg_image.dtype == np.uint8, "Rendered layer image must have dtype=np.uint8"
        assert fg_image.ndim == 3, "Rendered layer image must have 3 dimensions (H, W, C)"
        assert fg_image.shape[2] == 4, "Rendered layer image must have 4 channels (RGBA)"
        if key is not None:
            cache[key] = fg_image
        return fg_image

    def _composite(
        self, bg_image: np.ndarray, time: float,
        parent: tuple[int, int] = (0, 0),
        preview_level: int = 1,
        cache: Cache | None = None,
    ) -> np.ndarray:
        # Retrieve layer image
        layer_time = time - self.offset
        if layer_time < self.start_time or self.end_time <= layer_time:
            return bg_image
        fg_image = self._get_fg_image(time, cache)
        if fg_image is None:
            return bg_image

        # Get affine matrix and transform layer image
        p = self.transform.get_current_value(layer_time)
        result = _get_fixed_affine_matrix(fg_image, p, preview_level=preview_level)
        if result is None:
            return bg_image
        affine_matrix_fixed, (W, H), (offset_x, offset_y) = result
        fg_image_transformed = cv2.warpAffine(
            fg_image, affine_matrix_fixed, dsize=(W, H),
            flags=cv2.INTER_LINEAR, borderMode=cv2.BORDER_CONSTANT)

        # Composite bg_image and fg_image
        bg_image = alpha_composite(
            bg_image, fg_image_transformed,
            position=(offset_x - parent[0], offset_y - parent[1]),
            opacity=p.opacity, blending_mode=p.blending_mode)
        return bg_image

    def __call__(self, time: float) -> np.ndarray | None:
        layer_time = time - self.offset
        if not self.visible:
            return None
        frame = self.layer(layer_time)
        if frame is None:
            return None
        for effect in self._effects:
            frame = effect(frame, layer_time)
        return frame

    def _get_audio_data(self, start_time: float, end_time: float) -> tuple[float, float, np.ndarray] | None:
        if not self.audio:
            return None
        layer_time_start = max(self.start_time, start_time - self.offset)
        layer_time_end = min(self.end_time, end_time - self.offset)
        if layer_time_start >= layer_time_end:
            return None
        if not hasattr(self.layer, 'get_audio'):
            return None
        layer: AudioLayer = self.layer  # type: ignore
        audio = layer.get_audio(layer_time_start, layer_time_end)
        if audio is None:
            return None
        scale = _get_scale_by_block(
            self.audio_level, layer_time_start, audio.shape[1])
        return layer_time_start, layer_time_end, scale * audio

    def __repr__(self) -> str:
        return f"LayerItem(name={self.name!r}, layer={self.layer!r}, transform={self.transform!r}, " \
            f"offset={self.offset}, visible={self.visible})"


def _get_fixed_affine_matrix(
    fg_image: np.ndarray, p: TransformValue,
    preview_level: int = 1
) -> tuple[np.ndarray, tuple[int, int], tuple[int, int]] | None:
    h, w = fg_image.shape[:2]

    T1, SR = _get_T1(p), _get_SR(p)
    T2 = _get_T2(p, (w, h), p.origin_point)
    M = T1 @ SR @ T2
    P = np.array([
        [1 / preview_level, 0, 0],
        [0, 1 / preview_level, 0],
        [0, 0, 1]], dtype=np.float64)
    affine_matrix = (P @ M)[:2]

    corners_layer = np.array([
        [0, 0, 1],
        [0, h, 1],
        [w, 0, 1],
        [w, h, 1]], dtype=np.float64)
    corners_global = corners_layer @ affine_matrix.transpose()
    min_coords = np.ceil(corners_global.min(axis=0))
    max_coords = np.floor(corners_global.max(axis=0))
    WH = (max_coords - min_coords).astype(np.int32)
    W, H = WH[0], WH[1]
    if W == 0 or H == 0:
        return None
    offset_x, offset_y = int(min_coords[0]), int(min_coords[1])

    Pf = np.array([
        [1 / preview_level, 0, - offset_x],
        [0, 1 / preview_level, - offset_y],
        [0, 0, 1]], dtype=np.float64)
    affine_matrix_fixed = (Pf @ M)[:2]
    return affine_matrix_fixed, (W, H), (offset_x, offset_y)


def _get_T1(p: TransformValue) -> np.ndarray:
    return np.array([
        [1, 0, p.position[0] + p.anchor_point[0]],
        [0, 1, p.position[1] + p.anchor_point[1]],
        [0, 0, 1]], dtype=np.float64)


def _get_SR(p: TransformValue) -> np.ndarray:
    cos_t = np.cos((2 * np.pi * p.rotation) / 360)
    sin_t = np.sin((2 * np.pi * p.rotation) / 360)
    SR = np.array([
        [p.scale[0] * cos_t, - p.scale[0] * sin_t, 0],
        [p.scale[1] * sin_t, p.scale[1] * cos_t, 0],
        [0, 0, 1]], dtype=np.float64)
    return SR


def _get_T2(p: TransformValue, size: tuple[int, int], origin_point: Direction) -> np.ndarray:
    center_point = Direction.to_vector(
        origin_point, (float(size[0]), float(size[1])))
    T2 = np.array([
        [1, 0, - p.anchor_point[0] - center_point[0]],
        [0, 1, - p.anchor_point[1] - center_point[1]],
        [0, 0, 1]], dtype=np.float64)
    return T2


def _get_scale_by_block(audio_level: Attribute, start_time: float, n_samples: int) -> np.ndarray:
    n_blocks = (n_samples + AUDIO_BLOCK_SIZE - 1) // AUDIO_BLOCK_SIZE
    block_times = start_time + np.arange(n_blocks) * (AUDIO_BLOCK_SIZE / AUDIO_SAMPLING_RATE)
    block_level = audio_level.get_values(block_times)
    block_scale = 10.0 ** (block_level / 20.0)
    C = block_scale.shape[1]
    scale = np.broadcast_to(
        block_scale.transpose().reshape(C, n_blocks, 1),
        (C, n_blocks, AUDIO_BLOCK_SIZE)).reshape(C, n_blocks * AUDIO_BLOCK_SIZE)
    return scale[:, :n_samples]



from __future__ import annotations

from os import PathLike
from pathlib import Path
from typing import Sequence

import imageio
import librosa
import numpy as np
from PIL import Image as PILImage

from ..util import to_rgb
from .mixin import TimelineMixin
from .protocol import AUDIO_SAMPLING_RATE


class Image:
    """Still image layer to encapsulate various formats of image data and offer
    time-based keying.

    Args:
        img_file: the source of the image data. It can be a file path (``str`` or ``PathLike``),
            a `PIL.Image` object, or a two or three-dimensional ``numpy.ndarray`` with a shape of ``(H, W, C)``.
        duration: the duration for which the image should be displayed.
            Default is ``1000000.0`` (long enough time).
    """
    def __init__(
        self,
        img_file: str | PathLike | PILImage.Image | np.ndarray,
        duration: float = 1e6
    ) -> None:
        self._image: np.ndarray | None = None
        self._img_file: Path | None = None
        if isinstance(img_file, (str, PathLike)):
            self._img_file = Path(img_file)
            assert self._img_file.exists(), f"{self._img_file} does not exist"
        elif isinstance(img_file, PILImage.Image):
            image = np.asarray(img_file.convert("RGBA"))
            self._image = image
        elif isinstance(img_file, np.ndarray):
            assert img_file.dtype == np.uint8
            if img_file.ndim == 2:
                img = img_file[:, :, None]
                img = np.concatenate([
                    np.repeat(img, 3, axis=-1),
                    np.full_like(img, 255, dtype=np.uint8)],
                    axis=-1)
                self._image = img
            elif img_file.ndim == 3:
                if img_file.shape[2] == 3:
                    img = np.concatenate([
                        img_file,
                        np.full_like(img_file[:, :, :1], 255, dtype=np.uint8)],
                        axis=-1)
                    self._image = img
                elif img_file.shape[2] == 4:
                    self._image = img_file
                else:
                    raise ValueError(f"Invalid img_file shape: {img_file.shape}. Must be (H, W, 3) or (H, W, 4).")
                assert img_file.shape[2] == 4, "Image must have 4 channels (RGBA)"
            else:
                raise ValueError(f"Invalid img_file shape: {img_file.shape}")
        else:
            raise ValueError(f"Invalid img_file type: {type(img_file)}")

        self._duration = duration

    @property
    def image(self) -> np.ndarray | None:
        """The image data."""
        return self._read_image()

    @classmethod
    def from_color(clf, size: tuple[int, int], color: str | tuple[int, int, int], duration: float = 1e6) -> "Image":
        """Create a plain image with a given color.

        Args:
            size:
                the size of the image with a tuple of ``(width, height)``.
            color:
                the color of the image. It can be a color name (``str``) or a tuple of ``(R, G, B)``.
            duration:
                the duration for which the image should be displayed. Default is ``1000000.0`` (long enough time).

        Returns:
            An ``Image`` object.
        """
        assert size[0] > 0 and size[1] > 0
        rgb = np.array(to_rgb(color)).reshape(1, 1, 3)
        image = np.full((size[1], size[0], 4), 255, dtype=np.uint8)
        image[:, :, :3] = rgb
        return clf(image, duration=duration)

    @property
    def duration(self) -> float:
        """The duration for which the image should be displayed."""
        return self._duration

    @property
    def size(self) -> tuple[int, int]:
        """The size of the image with a tuple of `(width, height)`."""
        shape = self._read_image().shape[:2]
        return shape[1], shape[0]

    def get_key(self, time: float) -> bool:
        """Get the state index for the given time."""
        return 0 <= time < self.duration

    def _read_image(self) -> np.ndarray:
        if self._image is None:
            assert self._img_file is not None
            image = np.asarray(PILImage.open(self._img_file).convert("RGBA"))
            self._image = image
        return self._image

    def __call__(self, time: float) -> np.ndarray | None:
        if 0 <= time < self.duration:
            return self._read_image()
        return None


class ImageSequence(TimelineMixin):
    """Image sequence layer to encapsulate various formats of images.

    Args:
        start_times:
            a sequence of start times for each image.
        end_times:
            a sequence of end times for each image.
        img_files:
            a sequence of image data. Each element can be a file path (``str`` or ``Path``),
            a `PIL.Image` object, or a two or four-dimensional ``numpy.ndarray`` with a shape of ``(H, W, C)``.
    """

    @classmethod
    def from_files(
        cls,
        img_files: Sequence[str | PathLike | PILImage.Image | np.ndarray],
        each_duration: float = 1.0
    ) -> "ImageSequence":
        """Create an ``ImageSequence`` object from a sequence of image files.

        Different from ``ImageSequence.__init__``, this method does not require the start and end times,
        and the duration for each image is set to ``each_duration``.

        Args:
            img_files:
                a sequence of image data. Each element can be a file path (``str`` or ``PathLike``),
                a `PIL.Image` object, or a two or four-dimensional ``numpy.ndarray`` with a shape of ``(H, W, C)``.
            each_duration:
                the duration for which each image should be displayed. Default is ``1.0``.

        Returns:
            An ``ImageSequence`` object.
        """
        start_times = np.arange(len(img_files)) * each_duration
        end_times = start_times + each_duration
        return cls(start_times.tolist(), end_times.tolist(), img_files)

    @classmethod
    def from_dir(
        cls,
        img_dir: str | PathLike,
        each_duration: float = 1.0
    ) -> "ImageSequence":
        """Create an ``ImageSequence`` object from a directory of image files.

        Args:
            img_dir:
                a directory containing image files.
            each_duration:
                the duration for which each image should be displayed. Default is ``1.0``.

        Returns:
            An ``ImageSequence`` object.
        """
        img_dir = Path(img_dir)
        exts = set([".png", ".jpg", ".jpeg", ".bmp", ".tiff"])
        img_files = [
            p for p in sorted(img_dir.glob("*")) if p.is_file() and p.suffix in exts]
        return cls.from_files(img_files, each_duration)

    def __init__(
        self,
        start_times: Sequence[float],
        end_times: Sequence[float],
        img_files: Sequence[str | PathLike | PILImage.Image | np.ndarray]
    ) -> None:
        super().__init__(start_times, end_times)
        self.img_files = img_files
        self.images: list[np.ndarray | None] = [None] * len(img_files)
        for i, img_file in enumerate(img_files):
            if isinstance(img_file, (str, PathLike)):
                img_file = Path(img_file)
                assert Path(img_file).exists(), f"{img_file} does not exist"
            elif isinstance(img_file, PILImage.Image):
                self.images[i] = np.asarray(img_file.convert("RGBA"))
            elif isinstance(img_file, np.ndarray):
                self.images[i] = img_file
            else:
                raise ValueError(f"Invalid img_file type: {type(img_file)}")

    def get_key(self, time: float) -> int:
        """Get the state index for the given time."""
        idx = self.get_state(time)
        if idx < 0:
            return -1
        return idx

    def __call__(self, time: float) -> np.ndarray | None:
        idx = self.get_state(time)
        if idx < 0:
            return None
        if self.images[idx] is None:
            img_file = self.img_files[idx]
            assert isinstance(img_file, (str, PathLike))
            image = np.asarray(PILImage.open(str(img_file)).convert("RGBA"))
            self.images[idx] = image
        return self.images[idx]


class Video:
    """Video layer to encapsulate various formats of video data.

    Args:
        video_file:
            the source of the video data. It can be a file path (``str`` or ``Path``).
        audio:
            whether to include the audio layer. Default is ``True``.
    """

    def __init__(self, video_file: str | PathLike, audio: bool = True) -> None:
        self.video_file = Path(video_file)
        self._reader = imageio.get_reader(self.video_file)
        meta_data = self._reader.get_meta_data()
        self._fps = meta_data["fps"]
        self._size = meta_data["size"]
        self._n_frame = meta_data["nframes"]
        self._duration = meta_data["duration"]
        self._audio = audio
        self._audio_layer = None
        if audio and "audio_codec" in meta_data:
            self._audio_layer = Audio(video_file)

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_reader'] = None
        return state

    @property
    def fps(self) -> float:
        """The frame rate of the video."""
        return self._fps

    @property
    def size(self) -> tuple[int, int]:
        """The size of the video with a tuple of ``(width, height)``."""
        return self._size

    @property
    def n_frame(self) -> int:
        """The number of frames in the video."""
        return self._n_frame

    @property
    def duration(self) -> float:
        """The duration of the video."""
        return self._duration

    def has_audio(self) -> bool:
        """Return True if the video has audio layer."""
        return self._audio_layer is not None

    @property
    def audio(self) -> bool:
        """Whether the video has audio data."""
        return self._audio

    def get_key(self, time: float) -> int:
        """Get the state index for the given time."""
        if time < 0 or self.duration < time:
            return -1
        frame_index = int(time * self._fps)
        return frame_index

    def __call__(self, time: float) -> np.ndarray | None:
        if self._reader is None:
            self._reader = imageio.get_reader(self.video_file)
        frame_index = int(time * self._fps)
        try:
            frame = self._reader.get_data(frame_index)
        except IndexError:
            return None
        pil_frame = PILImage.fromarray(frame).convert("RGBA")
        return np.asarray(pil_frame)

    def get_audio(self, start_time: float, end_time: float) -> np.ndarray | None:
        if self._audio and self._audio_layer is not None:
            return self._audio_layer.get_audio(start_time, end_time)
        return None


class Audio:
    """Audio layer to encapsulate various formats of audio data.

    Args:
        audio_file:
            the source of the audio data. It can be a file path (``str`` or ``Path``)
            or a ``numpy.ndarray`` with a shape of ``(2, N)`` and a sampling rate of 44100.
            If the audio data is a mono channel, it will be broadcasted to a stereo channel.
    """

    def __init__(self, audio_file: str | PathLike | np.ndarray) -> None:
        self._audio_file: Path | None = None
        self._audio: np.ndarray | None = None
        self._duration: float | None = None
        if isinstance(audio_file, (str, PathLike)):
            self._audio_file = Path(audio_file)
            assert self._audio_file.exists(), f"{self._audio_file} does not exist"
        elif isinstance(audio_file, np.ndarray):
            if audio_file.dtype != np.float32:
                raise ValueError(f"Invalid audio_file dtype: {audio_file.dtype}. Must be np.float32")
            if audio_file.ndim == 1:
                audio = np.broadcast_to(audio_file[None, :], (2, len(audio_file)))
            elif audio_file.ndim == 2:
                assert audio_file.shape[0] == 2, "Audio must have 2 channels"
                audio = audio_file
            else:
                raise ValueError(f"Invalid audio_file shape: {audio_file.shape}")
            self._audio = audio
        else:
            raise ValueError(f"Invalid audio_file type: {type(audio_file)}")

    def _load_audio(self) -> np.ndarray:
        if self._audio is None:
            audio, _ = librosa.load(str(self._audio_file), sr=AUDIO_SAMPLING_RATE, mono=False)
            if audio.ndim == 1:
                audio = np.broadcast_to(audio[None, :], (2, len(audio)))
            self._audio = audio
        return self._audio

    @property
    def audio_file(self) -> Path | None:
        """The file path of the audio data."""
        return self._audio_file

    @property
    def duration(self) -> float:
        """The duration of the audio data."""
        if self._duration is not None:
            return self._duration

        if self._audio is not None:
            duration = self._audio.shape[1] / AUDIO_SAMPLING_RATE
            self._duration = duration
            return duration
        else:
            duration = librosa.get_duration(path=str(self._audio_file))
            self._duration = duration
            return duration

    @property
    def audio(self) -> np.ndarray:
        """The audio data with a shape of ``(2, N)`` with a sampling rate of 44100."""
        return self._load_audio()

    def __call__(self, time: float) -> np.ndarray | None:
        return None

    def get_key(self, time: float) -> int:
        """Get the state index for the given time.

        Note:
            This method always returns a constant value because the audio data does not affect the image data.
        """
        return 0

    def get_audio(self, start_time: float, end_time: float) -> np.ndarray | None:
        """Get the audio data for the given time range.

        Args:
            start_time: the start time of the audio data.
            end_time: the end time of the audio data.

        Returns:
            The audio data for the given time range. If no audio data is found, ``None`` is returned.
        """
        assert start_time < end_time, f"start_time ({start_time}) must be smaller than end_time ({end_time})"
        if end_time <= 0 or start_time >= self.duration:
            return None

        audio = self._load_audio()
        start_index = int(start_time * AUDIO_SAMPLING_RATE)
        end_index = int(end_time * AUDIO_SAMPLING_RATE)
        if start_index < 0:
            audio = np.pad(audio, ((0, 0), (-start_index, 0)))
            end_index = end_index - start_index
            start_index = 0
        if end_index > audio.shape[1]:
            audio = np.pad(audio, ((0, 0), (0, end_index - audio.shape[1])))
        return audio[:, start_index:end_index]


class AudioSequence:
    """Audio sequence layer to handle multiple audio files.

    Args:
        start_times:
            a sequence of start times for each audio.
        end_times:
            a sequence of end times for each audio.
        audio_files:
            a sequence of audio data. Each element can be a file path (``str`` or ``Path``)
            or a ``numpy.ndarray`` with a shape of ``(2, N)`` and a sampling rate of 44100.
            If the audio data is a mono channel, it will be broadcasted to a stereo channel.
        """

    def __init__(
        self,
        start_times: Sequence[float],
        end_times: Sequence[float],
        audio_files: Sequence[str | Path | np.ndarray],
    ) -> None:
        assert len(start_times) == len(end_times) == len(audio_files)
        assert np.all(np.asarray(start_times) < np.asarray(end_times))
        # Check if the time ranges are not overlapped
        assert np.all(np.asarray(end_times)[:-1] <= np.asarray(start_times)[1:])

        self.start_times = np.asarray(start_times, dtype=float)
        self.end_times = np.asarray(end_times, dtype=float)
        self.audio_files = list(audio_files)
        self._audio: list[np.ndarray | None] = [None] * len(audio_files)

    @property
    def duration(self) -> float:
        """The duration of the audio sequence."""
        return self.end_times[-1]

    def __call__(self, time: float) -> np.ndarray | None:
        return None

    def get_key(self, time: float) -> int:
        """Get the state index for the given time.

        Note:
            This method always returns a constant value because the audio data does not affect the image data."""
        return 0

    def _load_audio(self, index: int) -> np.ndarray:
        a = self._audio[index]
        if a is None:
            audio_file = self.audio_files[index]
            if isinstance(audio_file, (str, Path)):
                a_i, _ = librosa.load(audio_file, sr=AUDIO_SAMPLING_RATE, mono=False)
                if a_i.ndim == 1:
                    a_i = np.broadcast_to(a_i[None, :], (2, len(a_i)))
            elif isinstance(audio_file, np.ndarray):
                if audio_file.dtype != np.float32:
                    raise ValueError(f"Invalid audio_file dtype: {audio_file.dtype}. Must be np.float32")
                if audio_file.ndim == 1:
                    a_i = np.broadcast_to(audio_file[None, :], (2, len(audio_file)))
                elif audio_file.ndim == 2:
                    assert audio_file.shape[0] == 2, "Audio must have 2 channels"
                    a_i = audio_file
                else:
                    raise ValueError(f"Invalid audio_file shape: {audio_file.shape}")
            else:
                raise ValueError(f"Invalid audio_file type: {type(audio_file)}")
            self._audio[index] = a_i
            return a_i
        return a

    def get_audio(self, start_time: float, end_time: float) -> np.ndarray | None:
        """Get the audio data for the given time range.

        Args:
            start_time:
                the start time of the audio data.
            end_time:
                the end time of the audio data.

        Returns:
            The audio data for the given time range. If no audio data is found, ``None`` is returned.
        """
        assert start_time < end_time, f"start_time ({start_time}) must be smaller than end_time ({end_time})"
        if end_time <= 0 or start_time >= self.duration:
            return None
        w0 = int((end_time - start_time) * AUDIO_SAMPLING_RATE)
        if w0 <= 0:
            return None
        audio = np.zeros((2, w0), dtype=np.float32)

        start_index = max(
            0, int(np.searchsorted(self.start_times, start_time, side="right") - 1))
        end_index = min(
            len(self.end_times),
            int(np.searchsorted(self.end_times, end_time, side="left") + 1))
        if start_index == end_index:
            return None

        for i in range(start_index, end_index):
            ts = float(self.start_times[i])
            te = float(self.end_times[i])
            audio_i = self._load_audio(i)[:, :int((te - ts) * AUDIO_SAMPLING_RATE)]
            wi = audio_i.shape[1]

            p = int((ts - start_time) * AUDIO_SAMPLING_RATE)
            x1, x2 = max(0, p), - min(0, p)
            w = min(p + wi, w0) - x1
            if w <= 0:
                continue
            audio[:, x1:x1 + w] = audio_i[:, x2:x2 + w]
        return audio



from __future__ import annotations

import hashlib
import sys
import urllib.request
from os import PathLike
from pathlib import Path

import cv2
import numpy as np

from movis.util import to_rgb

try:
    import onnxruntime
    onnxruntime_available = True
except ImportError:
    onnxruntime_available = False


class ChromaKey:
    """An effect that extracts the foreground using the chroma key composition.

    It extracts the foreground by using the chroma key composition,
    which identifies the background by a given color.
    The color is specified by the lower and upper bounds in the HSV color space.

    Args:
        key_color:
            The key color in the RGB color space.
            It also supports the color name in CSS3 (e.g., ``blur`` and ``green``).
        color_space:
            The color space when extracting the background. Currently, only ``hsv`` is supported.
        key_color_range:
            The color range of the background.
            The range of each channel is ``[0, 360]``, ``[0, 1]``, and ``[0, 1]``
            if the color space is ``hsv``.
    """

    def __init__(
            self,
            key_color: tuple[int, int, int] | str = (0, 255, 0),
            color_space: str = 'hsv',
            key_color_range: tuple[float, float, float] = (20.0, 0.3, 0.3)):
        assert color_space == 'hsv', "Only HSV color space is supported."
        c = cv2.cvtColor(
            np.array(to_rgb(key_color), dtype=np.uint8).reshape(1, 1, 3), cv2.COLOR_RGB2HSV)[0, 0, :].astype(np.float64)
        self._key_color = c.astype(np.uint8)
        self._lower_color = np.clip(np.array([
            c[0] - key_color_range[0] / 2,
            c[1] - 255 * key_color_range[1],
            c[2] - 255 * key_color_range[2]]), 0, 255).astype(np.uint8)
        self._upper_color = np.clip(np.array([
            c[0] + key_color_range[0] / 2,
            c[1] + 255 * key_color_range[1],
            c[2] + 255 * key_color_range[2]]), 0, 255).astype(np.uint8)

    def __call__(self, prev_image: np.ndarray, time: float) -> np.ndarray:
        hsv_foreground = cv2.cvtColor(prev_image[:, :, :3], cv2.COLOR_RGB2HSV)
        mask = cv2.inRange(hsv_foreground, self._lower_color, self._upper_color)
        inverse_mask = cv2.bitwise_not(mask)
        alpha = inverse_mask
        new_img = np.concatenate([prev_image[:, :, :3], alpha[:, :, None]], axis=2).astype(np.uint8)
        return new_img


class RobustVideoMatting:
    """An effect that extracts the foreground using the RobustVideoMatting [Li2021].

    It uses a deep learning model to automatically identify the area of persons in a given frame
    and extract it as the foreground.

    .. note::
        This effect requires `onnxruntime <https://onnxruntime.ai/>`_ to be installed.

    .. note::
        While there is no need to set up greenbacks or
        other special imaging environments, the quality is not at the production level.
        This effect is useful in areas that generally do not require foreground extraction quality,
        such as presentation videos.

    Args:
        onnx_file:
            The path to the ONNX file of the model. Download it from https://github.com/PeterL1n/RobustVideoMatting
            and put it in an appropriate location. If None, the default model will be downloaded
            and chached in ``~/.cache/movis``. The default model is ``rvm_mobilenetv3_fp16.onnx``.
        downsample_ratio:
            The downsample ratio to accelerate the inference speed. The default value is 0.25.
        recurrent_state:
            The flag to indicate whether the model uses a reccurent state.
            Enabling this flag tends to improve quality, but may also produce unstable results in some cases.
            The default value is ``True``.
    """

    default_model_url = \
        'https://github.com/PeterL1n/RobustVideoMatting/releases/download/v1.0.0/rvm_mobilenetv3_fp16.onnx'
    default_md5_digest = 'f8c8ae3ed12f3d6ba09211b58a406a28'

    def __init__(
            self, onnx_file: str | PathLike | None = None, downsample_ratio: float = 0.25,
            recurrent_state: bool = True):
        if not onnxruntime_available:
            raise ImportError("onnxruntime is not installed")
        self._downsample_ratio = np.array([downsample_ratio], dtype=np.float32)
        self._onnx_file = Path(onnx_file) if onnx_file is not None else None
        self._session: onnxruntime.InferenceSession | None = None
        self._state: list[np.ndarray] = [np.zeros([1, 1, 1, 1], dtype=np.float16)] * 4
        self._recurrent_state = recurrent_state

    @property
    def recurrent_state(self) -> bool:
        """Whether the model uses a recurrent state."""
        return self._recurrent_state

    def get_key(self, time: float) -> float:
        """Return the key for caching."""
        if self._recurrent_state:
            return time
        else:
            return -1.0

    def __call__(self, prev_frame: np.ndarray, time: float) -> np.ndarray:
        if not onnxruntime_available:
            raise ImportError("onnxruntime is not installed")
        if self._session is None:
            if self._onnx_file is None:
                self._onnx_file = _download_and_cache(self.default_model_url, self.default_md5_digest)
            else:
                self._onnx_file = Path(self._onnx_file)
            self._session = onnxruntime.InferenceSession(str(self._onnx_file))

        img = prev_frame[:, :, :3].transpose(2, 0, 1)[None, :, :, :]
        img = img.astype(np.float16) / 255
        s0 = self._state
        _, alpha, s10, s11, s12, s13 = self._session.run([], {
            'src': img,
            'r1i': s0[0], 'r2i': s0[1], 'r3i': s0[2], 'r4i': s0[3],
            'downsample_ratio': self._downsample_ratio,
        })
        if self._recurrent_state:
            self._state = [s10, s11, s12, s13]
        alpha = np.clip(alpha * 255, 0, 255).astype(np.uint8)[0].transpose(1, 2, 0)
        dst = np.concatenate([prev_frame[:, :, :3], alpha], axis=2)
        return dst


def _download_and_cache(url, expected_md5: str | None = None) -> Path:
    filename = Path(url).name
    home_dir = Path.home()
    cache_dir = home_dir / ".cache" / "movis"
    cache_dir.mkdir(parents=True, exist_ok=True)
    cache_path = cache_dir / filename
    if cache_path.exists():
        actual_md5 = _calculate_md5(cache_path)
        if expected_md5 is None or actual_md5 == expected_md5:
            return cache_path
        else:
            sys.stderr.write(f"MD5 mismatch: expected {expected_md5}, got {actual_md5}.")
            sys.stderr.flush()

    sys.stderr.write(f'Downloading {filename} from {url}...\n')
    sys.stderr.flush()
    urllib.request.urlretrieve(url, cache_path)
    return cache_path


def _calculate_md5(file_path: Path) -> str:
    """Calculate the MD5 hash of a file."""
    with open(file_path, 'rb') as f:
        return hashlib.md5(f.read()).hexdigest()



import pathlib

from setuptools import find_packages, setup

BASE_DIR = pathlib.Path(__file__).resolve().parent


def parse_requirements(name: str = "requirements.txt"):
    with open(BASE_DIR / name) as f:
        return f.read().splitlines()


with open(BASE_DIR / "README.md", encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="movis",
    version="0.7",
    author="Masaki Saito",
    author_email="msaito@preferred.jp",
    description="A video editing library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="video video-processing video-editing",
    url="https://github.com/rezoo/movis",
    project_urls={
        "Source": "https://github.com/rezoo/movis",
    },
    license="MIT License",
    packages=find_packages(),
    include_package_data=True,
    package_data={
        "movis": ["py.typed"],
    },
    python_requires=">=3.9.0",
    install_requires=parse_requirements(),
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)



from __future__ import annotations

import cv2
import numpy as np
from PIL import Image
from PySide6.QtGui import QImage

from .enum import BlendingMode, MatteMode


def _normal(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return fg


def _blend_multiply(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return (bg.astype(np.int32) * fg.astype(np.int32) // 255).astype(np.uint8)


def _blend_overlay(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return np.where(bg < 128, 2 * bg * fg // 255, 255 - 2 * (255 - bg) * (255 - fg) // 255)


def _darken(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return np.minimum(bg, fg)


def _lighten(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return np.maximum(bg, fg)


def _blend_screen(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    x = 255 - (255 - bg).astype(np.int32) * (255 - fg).astype(np.int32) // 255
    return x.astype(np.uint8)


def _color_dodge(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    x = 255 * bg.astype(np.int32) / (255 - fg.astype(np.int32) + 1)
    return np.clip(x, 0, 255).astype(np.uint8)


def _color_burn(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    x = 255 * (255 - bg.astype(np.int32)) / (fg.astype(np.int32) + 1)
    return 255 - np.clip(x, 0, 255).astype(np.uint8)


def _linear_dodge(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return np.minimum(255, bg.astype(np.int32) + fg.astype(np.int32)).astype(np.uint8)


def _linear_burn(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return np.maximum(0, bg.astype(np.int32) + fg.astype(np.int32) - 255).astype(np.uint8)


def _hard_light(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return _blend_overlay(fg, bg)


def _blend_soft_light(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:

    def soft_light_dark(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
        return bg - (255 - 2 * fg) * bg * (255 - bg) // (255 ** 2)

    def soft_light_light(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
        def g_w3c(x: np.ndarray) -> np.ndarray:
            return np.where(
                x < 64,
                ((16 * x - (12 * 255 ** 2) * x + 4 * (255 ** 2)) * x) // (255 ** 2),
                np.sqrt(255 * x).astype(np.int32))

        return bg + (2 * fg - 255) * (g_w3c(bg) - bg) // 255

    return np.where(fg < 128, soft_light_dark(bg, fg), soft_light_light(bg, fg)).astype(np.uint8)


def _vivid_light(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    fg2 = 2 * fg.astype(np.int32)
    return np.where(
        fg > 127,
        _color_dodge(bg, 2 * (fg2 - 127)),
        _color_burn(bg, fg2))


def _linear_light(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    fg2 = 2 * fg.astype(np.int32)
    return np.where(
        fg > 127,
        _linear_dodge(bg, fg2 - 255),
        _linear_burn(bg, fg2)
    )


def _pin_light(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    fg2 = 2 * fg.astype(np.int32)
    return np.where(
        fg > 127,
        np.maximum(bg, fg2 - 255),
        np.minimum(bg, fg2)).astype(np.uint8)


def _difference(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return np.abs(bg.astype(np.int32) - fg.astype(np.int32)).astype(np.uint8)


def _exclusion(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    bg = bg.astype(np.int32)
    fg = fg.astype(np.int32)
    return np.clip(bg + fg - 2 * bg * fg // 255, 0, 255).astype(np.uint8)


def _subtract(bg: np.ndarray, fg: np.ndarray) -> np.ndarray:
    return np.clip(bg.astype(np.int32) - fg.astype(np.int32), 0, 255).astype(np.uint8)


BLENDING_MODE_TO_FUNC = {
    BlendingMode.NORMAL: _normal,
    BlendingMode.MULTIPLY: _blend_multiply,
    BlendingMode.SCREEN: _blend_screen,
    BlendingMode.OVERLAY: _blend_overlay,
    BlendingMode.DARKEN: _darken,
    BlendingMode.LIGHTEN: _lighten,
    BlendingMode.COLOR_DODGE: _color_dodge,
    BlendingMode.COLOR_BURN: _color_burn,
    BlendingMode.LINEAR_DODGE: _linear_dodge,
    BlendingMode.LINEAR_BURN: _linear_burn,
    BlendingMode.HARD_LIGHT: _hard_light,
    BlendingMode.SOFT_LIGHT: _blend_soft_light,
    BlendingMode.VIVID_LIGHT: _vivid_light,
    BlendingMode.LINEAR_LIGHT: _linear_light,
    BlendingMode.PIN_LIGHT: _pin_light,
    BlendingMode.DIFFERENCE: _difference,
    BlendingMode.EXCLUSION: _exclusion,
    BlendingMode.SUBTRACT: _subtract,
}


def _overlay(
        bg: np.ndarray, fg: np.ndarray, p_bg: tuple[int, int], p_fg: tuple[int, int],
        size: tuple[int, int], opacity: float,
        mode: BlendingMode = BlendingMode.NORMAL,
        matte_mode: MatteMode = MatteMode.NONE) -> np.ndarray:
    x_bg, y_bg = p_bg
    x_fg, y_fg = p_fg
    w, h = size
    bg_subset = bg[y_bg: (y_bg + h), x_bg: (x_bg + w)]
    fg_subset = fg[y_fg: (y_fg + h), x_fg: (x_fg + w)]

    if matte_mode == MatteMode.LUMINANCE:
        mask = cv2.cvtColor(bg_subset, cv2.COLOR_RGB2GRAY).astype(np.uint32)
        fg_a = (fg_subset[..., 3] * opacity).astype(np.uint32) \
            if opacity < 1.0 else fg_subset[..., 3].astype(np.uint32)
        bg[y_bg: (y_bg + h), x_bg: (x_bg + w), :3] = fg_subset[..., :3]
        bg[:, :, 3] = 0
        bg[y_bg: (y_bg + h), x_bg: (x_bg + w), 3] = (mask * fg_a) // 255
        return bg

    bg_subset = bg_subset.astype(np.uint32)
    fg_subset = fg_subset.astype(np.uint32)
    bg_a = bg_subset[..., 3:]
    fg_a = (fg_subset[..., 3:] * opacity).astype(np.uint32) if opacity < 1.0 else fg_subset[..., 3:]
    coeff1, coeff2 = 255 * fg_a, bg_a * (255 - fg_a)
    out_a = coeff1 + coeff2
    bg_rgb, fg_rgb = bg_subset[..., :3], fg_subset[..., :3]
    target_rgb = BLENDING_MODE_TO_FUNC[mode](bg_rgb, fg_rgb)
    out_rgb = (coeff1 * target_rgb + coeff2 * bg_rgb) // (out_a + (out_a == 0))
    bg[y_bg: (y_bg + h), x_bg: (x_bg + w), :3] = out_rgb.astype(np.uint8)
    if matte_mode == MatteMode.NONE:
        bg[y_bg: (y_bg + h), x_bg: (x_bg + w), 3:] = (out_a // 255).astype(np.uint8)
    elif matte_mode == MatteMode.ALPHA:
        pass  # Do nothing
    return bg


def _alpha_composite_numpy(
    bg_image: np.ndarray,
    fg_image: np.ndarray,
    position: tuple[int, int] = (0, 0),
    opacity: float = 1.0,
    blending_mode: BlendingMode = BlendingMode.NORMAL,
    matte_mode: MatteMode = MatteMode.NONE,
) -> np.ndarray:
    h1, w1 = bg_image.shape[:2]
    h2, w2 = fg_image.shape[:2]

    x1, y1 = max(0, position[0]), max(0, position[1])
    x2, y2 = - min(0, position[0]), - min(0, position[1])
    w = min(position[0] + w2, w1) - x1
    h = min(position[1] + h2, h1) - y1

    if w <= 0 or h <= 0:
        return bg_image

    return _overlay(
        bg_image, fg_image, (x1, y1), (x2, y2), (w, h), opacity,
        blending_mode, matte_mode)


def _alpha_composite_pil(
    bg_image: np.ndarray,
    fg_image: np.ndarray,
    position: tuple[int, int] = (0, 0),
    opacity: float = 1.0,
) -> np.ndarray:
    assert 0.0 <= opacity <= 1.0, f"opacity must be in [0, 1], but {opacity} is given."
    if opacity < 1.0:
        fg_image = fg_image.copy()
        c_alpha = fg_image[:, :, 3].astype(np.uint16)
        a = int(np.round(opacity * 255))
        c_alpha = (c_alpha * a // 255).astype(np.uint8)
        fg_image[:, :, 3] = c_alpha
    base_img_pil = Image.fromarray(bg_image)
    base_img_pil.alpha_composite(
        Image.fromarray(fg_image), (position[0], position[1]))
    return np.asarray(base_img_pil)


def alpha_composite(
    bg_image: np.ndarray,
    fg_image: np.ndarray,
    position: tuple[int, int] = (0, 0),
    opacity: float = 1.0,
    blending_mode: str | BlendingMode = BlendingMode.NORMAL,
    matte_mode: MatteMode = MatteMode.NONE,
) -> np.ndarray:
    """Perform alpha compositing of two images (with alpha channels).

    This function asserts that both the background and foreground images have 4 channels (RGBA) and
    ``dtype=numpy.uint8``. If the background image is not writeable, a copy will be made.

    Args:
        bg_image:
            The background image as a 3D numpy array of shape ``(height, width, 4)``.
            The image should have 4 channels (RGBA), with ``dtype=numpy.uint8``.
        fg_image:
            the foreground image as a 3D numpy array of shape ``(height, width, 4)``.
            The image should have 4 channels (RGBA), with ``dtype=numpy.uint8``.
        position:
            The x, y coordinates indicating where the top-left corner of the
            foreground image should be placed on the background image.
            Default is ``(0, 0)``.
        opacity:
            The opacity level of the foreground image, between 0.0 and 1.0.
            Default is 1.0.
        blending_mode:
            The blending mode used for compositing the two images.
            Available modes are defined in the ``BlendingMode`` enum.
            Default is ``BlendingMode.NORMAL``. Note that the blending mode can also be
            specified as a string.
        matte_mode:
            The mode used for handling the matte channel.
            Available modes are defined in the ``MatteMode`` enum (``NONE``, ``ALPHA``, and ``LUMINANCE``).
            Default is ``MatteMode.NONE``. Note that the matte mode can also be specified as a string.

    Returns:
        The composited image as a 3D numpy array of shape ``(height, width, 4)``
        with ``dtype=numpy.uint8``.
    """
    assert bg_image.ndim == 3
    assert fg_image.ndim == 3
    assert bg_image.shape[2] == 4
    assert fg_image.shape[2] == 4
    assert bg_image.dtype == np.uint8
    assert fg_image.dtype == np.uint8
    if not bg_image.flags.writeable:
        bg_image = bg_image.copy()
    if blending_mode == BlendingMode.NORMAL and matte_mode == MatteMode.NONE:
        # Use PIL for normal blending mode
        # because it is faster than my implementation
        return _alpha_composite_pil(bg_image, fg_image, position, opacity)
    else:
        mode = BlendingMode.from_string(blending_mode) \
            if isinstance(blending_mode, str) else blending_mode
        return _alpha_composite_numpy(
            bg_image, fg_image, position, opacity, mode, matte_mode)


def qimage_to_numpy(image: QImage) -> np.ndarray:
    """Convert a QImage to a numpy ndarray.

    .. note::
        It asserts that the input QImage format is ``QImage.Format.Format_ARGB32``.
        The memory layout of the returned numpy array corresponds to the QImage layout.

    Args:
        image:
            The input QImage object. The function assumes that the image format
            is ``QImage.Format.Format_ARGB32``.

    Returns:
        A 3D numpy array of shape ``(height, width, 4)`` representing
        ``QImage``. The returned array will have 4 channels (RGBA) and dtype will be uint8.
    """
    assert image.format() == QImage.Format.Format_ARGB32
    ptr = image.bits()
    array_shape = (image.height(), image.width(), 4)
    return np.array(ptr).reshape(array_shape)



from __future__ import annotations

from typing import Sequence

import numpy as np

from movis.enum import Direction
from movis.layer.composition import Composition
from movis.layer.protocol import BasicLayer


def concatenate(layers: Sequence[BasicLayer], size: tuple[int, int] | None = None) -> Composition:
    """Concatenate layers into a single composition.

    Args:
        layers:
            Layers to concatenate.
        size:
            Size of the composition. If ``None``, the size of the layer is estimated.

    Returns:
        Composition with all layers concatenated.

    Examples:
        >>> import movis as mv
        >>> layer1 = mv.layer.Image("image1.png", duration=1.0)
        >>> layer2 = mv.layer.Image("image2.png", duration=2.0)
        >>> composition = mv.concatenate([layer1, layer2])  # concatenate two layers
        >>> composition.duration
        3.0
    """
    if size is None:
        img = layers[0](0.0)
        if img is None:
            raise ValueError("Cannot determine size of composition.")
        size = img.shape[1], img.shape[0]
    duration = sum(layer.duration for layer in layers)
    composition = Composition(size=size, duration=duration)
    time = 0.0
    for layer in layers:
        composition.add_layer(layer, offset=time)
        time += layer.duration
    return composition


def repeat(layer: BasicLayer, n_repeat: int, size: tuple[int, int] | None = None) -> Composition:
    """Repeat a layer multiple times.

    Args:
        layer:
            Layer to repeat.
        n_repeat:
            Number of times to repeat the layer.
        size:
            Size of the composition. If ``None``, the size of the layer is estimated.

    Returns:
        Composition with the layer repeated.

    Examples:
        >>> import movis as mv
        >>> layer = mv.layer.Image("image.png", duration=1.0)
        >>> composition = mv.repeat(layer, 3)  # repeat 3 times
        >>> composition.duration
        3.0
    """
    if size is None:
        img = layer(0.0)
        if img is None:
            raise ValueError("Cannot determine size of composition.")
        size = img.shape[1], img.shape[0]
    duration = layer.duration * n_repeat
    composition = Composition(size=size, duration=duration)
    for i in range(n_repeat):
        composition.add_layer(layer, offset=i * layer.duration)
    return composition


def trim(
    layer: BasicLayer, start_times: Sequence[float], end_times: Sequence[float],
    size: tuple[int, int] | None = None
) -> Composition:
    """Trim a layer with given time intervals and concatenate them.

    Args:
        layer:
            Layer to trim.
        start_times:
            Start times of the intervals.
        end_times:
            End times of the intervals.
        size:
            Size of the composition. If ``None``, the size of the layer is estimated.

    Returns:
        Composition with the layer trimmed and concatenated.

    Examples:
        >>> import movis as mv
        >>> layer = mv.layer.Video("video.mp4")
        >>> composition = mv.trim(layer, [0.0, 2.0], [1.0, 3.0])  # select 0.0-1.0 and 2.0-3.0, and concatenate them
        >>> composition.duration
        2.0
    """
    assert 0 < len(start_times)
    assert len(start_times) == len(end_times)
    starts = np.array(start_times, dtype=np.float64)
    ends = np.array(end_times, dtype=np.float64)
    assert np.all(starts < ends)
    if size is None:
        img = layer(0.0)
        if img is None:
            raise ValueError("Cannot determine size of composition.")
        size = img.shape[1], img.shape[0]
    durations = ends - starts
    total_duration = float(durations.sum())
    offsets = np.cumsum(np.concatenate([[0.], durations]))[:-1] - starts

    composition = Composition(size=size, duration=total_duration)
    for start, end, offset in zip(starts, ends, offsets):
        composition.add_layer(layer, offset=offset, start_time=start, end_time=end)
    return composition


def tile(
    layers: Sequence[BasicLayer], rows: int, cols: int,
    size: tuple[int, int] | None = None,
) -> Composition:
    """Tile layers into a single composition.

    Args:
        layers:
            Layers to tile. Note that the order of the layers is row-major.
            For example, if ``layers`` is ``[a, b, c, d]`` and
            ``rows`` and ``cols`` are both ``2``, the composition will be: ``[[a, b], [c, d]]``.
        rows:
            Number of rows.
        cols:
            Number of columns.
        size:
            Size of each layer. Note that ``tile`` assumes that all layers have the same size.
            If ``None``, the size of the layer is estimated.

    .. note::
        The layer resolution specified in ``size`` does not have to be the actual layer resolution.
        For example, if ``size`` is specified to be larger than the actual layer size,
        each layer is placed in the center of each tile.

    Returns:
        Composition with all layers tiled. The size is ``(cols * w, rows * h)``,
        where ``w`` and ``h`` are the width and height of each layer, respectively.

    Examples:
        >>> import movis as mv
        >>> import numpy as np
        >>> layer1 = mv.layer.Image(np.zeros((100, 100, 4), dtype=np.uint8), duration=1.0)
        >>> layer2 = mv.layer.Image(np.zeros((100, 100, 4), dtype=np.uint8), duration=1.0)
        >>> composition = mv.tile([layer1, layer2], rows=1, cols=2)  # tile 1x2
        >>> composition.duration
        1.0
        >>> composition.size
        (200, 100)
    """
    assert len(layers) == rows * cols, \
        f"Number of layers ({len(layers)}) must be equal to rows * cols ({rows * cols})."
    if size is None:
        result = layers[0](0.0)
        if result is None:
            raise ValueError("Cannot determine size of composition.")
        w, h = result.shape[1], result.shape[0]
    else:
        w, h = size

    W = cols * w
    H = rows * h
    duration = max(layer.duration for layer in layers)
    composition = Composition(size=(W, H), duration=duration)
    for i in range(rows):
        for j in range(cols):
            x = (j + 0.5) * w
            y = (i + 0.5) * h
            composition.add_layer(
                layers[i * cols + j], position=(x, y))
    return composition


def crop(layer: BasicLayer, rect: tuple[int, int, int, int]) -> Composition:
    """Crop a layer from a specified rectangle.

    Args:
        layer:
            Layer to crop.
        rect:
            Rectangle to crop. (x, y, width, height)

    Returns:
        Composition with the layer cropped.

    Examples:
        >>> import movis as mv
        >>> layer = mv.layer.Image("image.png", duration=1.0)
        >>> composition = mv.crop(layer, (10, 20, 100, 200))  # crop from (10, 20) with size (100, 200)
        >>> composition.duration
        1.0
        >>> composition.size
        (100, 200)
    """
    assert len(rect) == 4
    x, y, w, h = rect
    duration = layer.duration
    composition = Composition(size=(w, h), duration=duration)
    composition.add_layer(
        layer, position=(-x, -y), origin_point=Direction.TOP_LEFT)
    return composition


def switch(
    layers: Sequence[BasicLayer],
    start_times: Sequence[float], cams: Sequence[int],
    size: tuple[int, int] | None = None, duration: float | None = None
) -> Composition:
    """Switch layers at specified times.

    Args:
        layers:
            Layers to switch.
        start_times:
            Start times of the intervals.
        cams:
            Scene numbers of the intervals.
        size:
            Size of the composition. If ``None``, the size of the layer is estimated.
        duration:
            Duration of the composition. If ``None``, the duration is estimated.

    Returns:
        Composition with the layer switched.

    Examples:
        >>> import movis as mv
        >>> layer1 = mv.layer.Image("image1.png", duration=5.0)
        >>> layer2 = mv.layer.Image("image2.png", duration=5.0)
        >>> # Show layer1 at 0.0-2.0 and layer2 at 2.0-5.0
        >>> composition = mv.switch([layer1, layer2], [0.0, 2.0], [0, 1])
        >>> composition.duration
        5.0
    """
    assert len(start_times) == len(cams)
    assert 0 <= min(cams)
    assert max(cams) < len(layers)
    starts = np.array(start_times, dtype=np.float64)
    assert np.all(starts[1:] > starts[:-1])
    if size is None:
        img = layers[0](0.0)
        if img is None:
            raise ValueError("Cannot determine size of composition.")
        size = img.shape[1], img.shape[0]
    if duration is None:
        duration = min(layer.duration for layer in layers)
    composition = Composition(size=size, duration=duration)
    times = np.concatenate([starts, [duration]])
    for start_time, end_time, ind in zip(times[:-1], times[1:], cams):
        composition.add_layer(
            layers[ind], start_time=start_time, end_time=end_time)
    return composition


def insert(
    source: BasicLayer, target: BasicLayer, time: float,
    size: tuple[int, int] | None = None
) -> Composition:
    """Insert a target layer into a source layer at a specified time.

    For instance, consider inserting a brief eye-catch scene to indicate a change in chapters within
    a long interview video. ``mv.insert()`` is used for such purposes, to insert a short scene into a longer one:

    |------------------------|    |------------|----------|------------|
    |    Source layer        | -> |   Source   |  target  |   Source   |
    |------------------------|    |------------|----------|------------|

    Args:
        source:
            The layer to insert the target layer into.
        target:
            The layer to insert.
        time:
            The time to insert the target layer.
        size:
            Size of the returned composition. If ``None``, the size of the source layer is used.

    Returns:
        Composition with the target layer inserted.

    Examples:
        >>> import movis as mv
        >>> source = mv.layer.Image("source.png", duration=5.0)
        >>> target = mv.layer.Image("target.png", duration=1.0)
        >>> composition = mv.insert(source, target, time=2.0)
        >>> composition.duration
        6.0
        >>> composition(0.0)  # source layer
        >>> composition(2.0)  # target layer
        >>> composition(3.0)  # source layer
    """
    if size is None:
        img = source(0.0)
        if img is None:
            raise ValueError("Cannot determine size of composition.")
        size = img.shape[1], img.shape[0]
    duration = source.duration + target.duration
    composition = Composition(size=size, duration=duration)
    composition.add_layer(source, end_time=time)
    composition.add_layer(target, offset=time)
    composition.add_layer(source, offset=target.duration, start_time=time)
    return composition



from __future__ import annotations

import sys
from typing import Any, Callable, Hashable, Sequence

import numpy as np
from PySide6.QtCore import QCoreApplication, QPointF, QRectF, Qt
from PySide6.QtGui import (QBrush, QColor, QFont, QFontDatabase, QFontMetrics,
                           QImage, QPainter, QPainterPath, QPen)
from PySide6.QtWidgets import QApplication

from movis.imgproc import qimage_to_numpy

from ..attribute import Attribute, AttributesMixin, AttributeType
from ..enum import TextAlignment
from ..util import to_rgb
from .mixin import TimelineMixin


class FillProperty:
    """A property for filling a shape.

    Args:
        color:
            The color of the shape with a tuple of ``(r, g, b)``
            or a string representing a color name (`e.g.,` ``"#ff0000" or "red"``).
        opacify:
            The opacity of the shape in the range of ``[0, 1]``.
    """

    def __init__(self, color: tuple[int, int, int] | str, opacify: float = 1.):
        self._color: tuple[int, int, int] = to_rgb(color)
        self._opacity: float = float(opacify)

    @property
    def color(self) -> tuple[int, int, int]:
        return self._color

    @property
    def opacity(self) -> float:
        return self._opacity


class StrokeProperty:
    """A property for stroking a shape.

    Args:
        color:
            The color of the shape with a tuple of ``(r, g, b)``
            or a string representing a color name (`e.g.,` ``"#ff0000" or "red"``).
        width:
            The width of the stroke.
        opacity:
            The opacity of the shape in the range of ``[0, 1]``.
    """

    def __init__(self, color: tuple[int, int, int] | str, width: float = 1., opacity: float = 1.):
        self._color: tuple[int, int, int] = to_rgb(color)
        self._width: float = float(width)
        self._opacity: float = float(opacity)

    @property
    def color(self) -> tuple[int, int, int]:
        return self._color

    @property
    def width(self) -> float:
        return self._width

    @property
    def opacity(self) -> float:
        return self._opacity


class Line(AttributesMixin):
    """Draw a line from ``start`` to ``end``.

    Args:
        size:
            The size of the canvas with a tuple of ``(width, height)``.
        start:
            The start point of the line with a tuple of ``(x, y)``.
        end:
            The end point of the line with a tuple of ``(x, y)``.
        color:
            The color of the line with a tuple of ``(r, g, b)``
            or a string representing a color name (`e.g.,` ``"#ff0000" or "red"``).
        width:
            The width of the line.
        duration:
            The duration for which the line should be displayed.

    Animateable Attributes:
        ``trim_start``
            The start point of the line to be drawn in the range of ``[0, 1]``.
            The default value is ``0``.
        ``trim_end``
            The end point of the line to be drawn in the range of ``[0, 1]``.
            The default value is ``1``.
        ``start`` ``end`` ``color`` ``width``
            These attributes can be animated as well.
    """
    def __init__(
        self,
        size: tuple[int, int] = (100, 100),
        start: tuple[float, float] | np.ndarray = (0., 0.),
        end: tuple[float, float] | np.ndarray = (100., 100.),
        color: tuple[int, int, int] | str = (255, 255, 255),
        width: float = 1.,
        duration: float = 1e6,
    ) -> None:
        self.size = size
        self.start = Attribute(start, value_type=AttributeType.VECTOR2D)
        self.end = Attribute(end, value_type=AttributeType.VECTOR2D)
        self.color = Attribute(to_rgb(color), value_type=AttributeType.COLOR, range=(0., 255.))
        self.width = Attribute(width, value_type=AttributeType.SCALAR, range=(0., 1e6))
        self.trim_start = Attribute(0., value_type=AttributeType.SCALAR, range=(0., 1.))
        self.trim_end = Attribute(1., value_type=AttributeType.SCALAR, range=(0., 1.))
        self._duration = duration

    @property
    def duration(self) -> float:
        return self._duration

    def __call__(self, time: float) -> np.ndarray | None:
        p0 = self.start(time)
        p1 = self.end(time)
        trim_start = self.trim_start(time)[0]
        trim_end = self.trim_end(time)[0]
        assert 0. <= trim_start <= trim_end <= 1.
        p_start = p0 + trim_start * (p1 - p0)
        p_end = p0 + trim_end * (p1 - p0)
        r, g, b = tuple(np.round(self.color(time)).astype(int))

        W, H = self.size
        image = QImage(W, H, QImage.Format.Format_ARGB32)
        image.fill(QColor(0, 0, 0, 0))

        painter = QPainter(image)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        painter.setPen(QPen(QColor(b, g, r, 255), self.width(time)[0]))
        painter.drawLine(QPointF(*p_start), QPointF(*p_end))

        painter.end()
        return qimage_to_numpy(image)


class Rectangle(AttributesMixin):
    """Draw a rectangle with rounded corners.

    Args:
        size:
            The size of the rectangle with a tuple of ``(width, height)``.
        radius:
            The radius of the rounded corners. The default value is ``0``.
        color:
            The color of the rectangle with a tuple of ``(r, g, b)`` or a string (e.g., ``"#ff0000" or "red"``).
            If ``None``, this layer uses the ``contents`` argument to draw the rectangle.
        contents:
            A sequence of ``FillProperty`` or ``StrokeProperty`` objects,
            which will be drawn on top of the rectangle.
            If an empty sequence is given, the rectangle will be filled with the ``color`` argument.
        duration:
            The duration for which the rectangle should be displayed.

    Animateable Attributes:
        ``size``
        ``radius``
    """
    def __init__(
        self,
        size: tuple[float, float] = (100., 100.),
        radius: float = 0.,
        color: tuple[int, int, int] | str | None = None,
        contents: Sequence[FillProperty | StrokeProperty] = (),
        duration: float = 1e6
    ) -> None:
        self.size = Attribute(size, value_type=AttributeType.VECTOR2D, range=(0., 1e6))
        self.radius = Attribute(radius, value_type=AttributeType.SCALAR, range=(0., 1e6,))
        if color is None:
            self.contents = contents
        else:
            self.contents = (FillProperty(color=to_rgb(color)),)
        self._duration = duration

    @property
    def duration(self) -> float:
        return self._duration

    def __call__(self, time: float) -> np.ndarray | None:
        if len(self.contents) == 0:
            return None
        size = [float(x) for x in self.size(time)]
        w, h = float(size[0]), float(size[1])
        radius = float(self.radius(time))

        eps = 1.
        max_stroke = _get_max_stroke(self.contents)
        W = np.floor(w + max_stroke + 2 * eps)
        H = np.floor(h + max_stroke + 2 * eps)
        image = QImage(W, H, QImage.Format.Format_ARGB32)
        max_color = _get_max_color(self.contents)
        if max_color is None:
            image.fill(QColor(0, 0, 0, 0))
        else:
            image.fill(QColor(*max_color, 0))

        painter = QPainter(image)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        rect = QRectF(eps + max_stroke / 2, eps + max_stroke / 2, w, h)
        for c in self.contents:
            if isinstance(c, FillProperty):
                r, g, b = c.color
                a = round(255 * c.opacity)
                painter.setBrush(QBrush(QColor(b, g, r, a)))
                painter.setPen(QPen(QColor(b, g, r, 0)))
                painter.drawRoundedRect(rect, radius, radius, mode=Qt.SizeMode.AbsoluteSize)
            elif isinstance(c, StrokeProperty):
                r, g, b = c.color
                a = round(255 * c.opacity)
                painter_path = QPainterPath()
                painter_path.addRoundedRect(rect, radius, radius, mode=Qt.SizeMode.AbsoluteSize)
                painter.setBrush(QBrush(QColor(b, g, r, 0)))
                painter.setPen(QPen(QColor(b, g, r, a), c.width))
                painter.drawPath(painter_path)
            else:
                raise ValueError(f"Invalid content type: {type(c)}")
        painter.end()
        return qimage_to_numpy(image)


class Ellipse(AttributesMixin):
    """Draw an ellipse.

    Args:
        size:
            The size of the ellipse with a tuple of ``(width, height)``.
        color:
            The color of the ellipse with a tuple of ``(r, g, b)``,
            or a string representing a color name (`e.g.,` ``"#ff0000" or "red"``).
            If ``None``, this layer uses the ``contents`` argument to draw the ellipse.
        contents:
            A sequence of ``FillProperty`` or ``StrokeProperty`` objects,
            which will be drawn on top of the ellipse.
            If an empty sequence is given, the ellipse will be filled with the ``color`` argument.
        duration:
            The duration for which the ellipse should be displayed.

    Animateable Attributes:
        ``size``
    """
    def __init__(
        self,
        size: tuple[float, float] = (100., 100.),
        color: tuple[int, int, int] | str | None = None,
        contents: Sequence[FillProperty | StrokeProperty] = (),
        duration: float = 1e6
    ) -> None:
        self.size = Attribute(size, value_type=AttributeType.VECTOR2D, range=(0., 1e6))
        if color is None:
            self.contents = contents
        else:
            self.contents = (FillProperty(color=to_rgb(color)),)
        self._duration = duration

    @property
    def duration(self) -> float:
        return self._duration

    def __call__(self, time: float) -> np.ndarray | None:
        if len(self.contents) == 0:
            return None
        size = [float(x) for x in self.size(time)]
        w, h = float(size[0]), float(size[1])

        eps = 1.
        max_stroke = _get_max_stroke(self.contents)
        W = np.floor(w + max_stroke + 2 * eps)
        H = np.floor(h + max_stroke + 2 * eps)
        image = QImage(W, H, QImage.Format.Format_ARGB32)
        max_color = _get_max_color(self.contents)
        if max_color is None:
            image.fill(QColor(0, 0, 0, 0))
        else:
            image.fill(QColor(*max_color, 0))

        painter = QPainter(image)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        rect = QRectF(eps + max_stroke / 2, eps + max_stroke / 2, w, h)
        for c in self.contents:
            if isinstance(c, FillProperty):
                r, g, b = c.color
                a = round(255 * c.opacity)
                painter.setBrush(QBrush(QColor(b, g, r, a)))
                painter.setPen(QPen(QColor(b, g, r, 0)))
                painter.drawEllipse(rect)
            elif isinstance(c, StrokeProperty):
                r, g, b = c.color
                a = round(255 * c.opacity)
                painter_path = QPainterPath()
                painter_path.addEllipse(rect)
                painter.setPen(QPen(QColor(b, g, r, a), c.width))
                painter.setBrush(QBrush(QColor(b, g, r, 0)))
                painter.drawPath(painter_path)
            else:
                raise ValueError(f"Invalid content type: {type(c)}")
        painter.end()
        return qimage_to_numpy(image)


class _TextWithTime(TimelineMixin):
    def __init__(self, start_times: Sequence[float], end_times: Sequence[float], texts: Sequence[str]):
        super().__init__(start_times, end_times)
        self.texts = texts

    def __call__(self, time: float) -> str:
        idx = self.get_state(time)
        if idx >= 0:
            return self.texts[idx]
        else:
            return ''


class Text(AttributesMixin):
    """Draw a text.

    Args:
        text:
            the text to be drawn. It can be a string or a callable object.
            If it is a callable object, it must accept a float value representing the time,
            and return a string.
        font_size:
            the font size of the text.
        font_family:
            the font family of the text. It must be one of the available fonts (`e.g.,` ``"Helvetica"``).
            To see the list of available fonts, run ``movis.layer.Text.available_fonts()``.
        font_style:
            the font style of the text. It must be one of the available styles of the given font family
            (`e.g.,` ``"Bold"``). To see the list of available styles,
            un ``movis.layer.Text.available_styles(font_name)``.
        color:
            the color of the text with a tuple of ``(r, g, b)``,
            or a string representing a color name (`e.g.,` ``"#ff0000" or "red"``).
            If ``None``, this layer uses the ``contents`` argument to draw the text.
        contents:
            A sequence of ``FillProperty`` or ``StrokeProperty`` objects,
            which will be drawn on top of the text.
            If an empty sequence is given, the text will be filled with the ``color`` argument.
        line_spacing:
            the line spacing of the text. If ``None``, the line spacing is automatically determined.
        text_alignment:
            the text alignment. If string is given, it must be one of ``"left"``, ``"center"``, or ``"right"``.
            It also accepts a ``TextAlignment`` enum.
        duration:
            the duration for which the text should be displayed.

        Animateable Attributes:
            ``font_size``
    """

    @staticmethod
    def available_fonts() -> Sequence[str]:
        """Returns the list of available fonts."""
        if QCoreApplication.instance() is None:
            QApplication(sys.argv[:1])
        return QFontDatabase.families()

    @staticmethod
    def available_styles(font_name: str) -> Sequence[str]:
        """Returns the list of available styles of the given font family."""
        if QCoreApplication.instance() is None:
            QApplication(sys.argv[:1])
        return QFontDatabase.styles(font_name)

    @classmethod
    def from_timeline(
        cls,
        start_times: Sequence[float],
        end_times: Sequence[float],
        texts: Sequence[str],
        **kwargs
    ) -> 'Text':
        """Create a text layer from a timeline.

        This method is useful when you want to display different texts at different times
        (e.g., displaying a subtitle). Note that other arguments are the same as the constructor.

        Args:
            start_times:
                A sequence of start times of the texts.
            end_times:
                A sequence of end times of the texts.
            texts:
                A sequence of texts to be displayed.

        Returns:
            A new `Text` object.
        """
        assert len(start_times) == len(texts)

        kwargs['duration'] = max(end_times)
        return cls(text=_TextWithTime(start_times, end_times, texts), **kwargs)

    def __init__(
        self,
        text: str | Callable[[float], str],
        font_size: float,
        font_family: str = 'Sans Serif',
        font_style: str | None = None,
        color: tuple[int, int, int] | str | None = None,
        contents: Sequence[FillProperty | StrokeProperty] = (),
        line_spacing: int | None = None,
        text_alignment: TextAlignment | str = TextAlignment.CENTER,
        duration: float = 1e6
    ) -> None:
        self._text = text
        self._font_family = font_family
        self._font_style = font_style
        self.font_size = Attribute(font_size, value_type=AttributeType.SCALAR, range=(0., 1e6))
        if color is None:
            self._contents = tuple(contents)
        else:
            self._contents = (FillProperty(color=to_rgb(color)),)
        self._line_spacing = line_spacing
        self._text_alignment = TextAlignment.from_string(text_alignment) \
            if isinstance(text_alignment, str) else text_alignment
        self._duration = duration
        if QCoreApplication.instance() is None:
            QApplication(sys.argv[:1])
        self._init_app = True

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        state['_init_app'] = False
        return state

    @property
    def text(self) -> str | Callable[[float], str]:
        return self._text

    @property
    def font_family(self) -> str:
        return self._font_family

    @property
    def font_style(self) -> str | None:
        return self._font_style

    @property
    def contents(self) -> tuple[FillProperty | StrokeProperty, ...]:
        return self._contents

    @property
    def line_spacing(self) -> int | None:
        return self._line_spacing

    @property
    def text_alignment(self) -> TextAlignment:
        return self._text_alignment

    @property
    def duration(self) -> float:
        return self._duration

    def get_text(self, time: float = 0.) -> str:
        """Returns the text to be drawn at the given time."""
        if isinstance(self.text, str):
            return self.text
        elif callable(self.text):
            return self.text(time)
        else:
            raise ValueError(f"Invalid text type: {type(self.text)}")

    def _get_qfont(self, time: float) -> QFont:
        if self.font_style is None:
            return QFont(self.font_family, round(float(self.font_size(time))))
        else:
            return QFontDatabase.font(
                self.font_family, self.font_style, round(float(self.font_size(time))))

    def get_size(self, time: float = 0.) -> tuple[int, int]:
        """Returns the size of the text at the given time.

        .. note::
            The returned size is the size of the text drawn on the canvas,
            and it may be different from the size of the text itself.

        Args:
            time:
                The time at which the size of the text is measured.
        """
        qfont = self._get_qfont(time)
        metrics = QFontMetrics(qfont)
        text = self.get_text(time)
        lines = text.split('\n')
        W, H = 0, 0
        for i, line in enumerate(lines):
            rect = metrics.boundingRect(line)
            W = max(W, rect.width() + rect.x())
            if self.line_spacing is None or i == len(lines) - 1:
                H += (rect.height() - rect.y())
            else:
                H += self.line_spacing
        return (W, H)

    def get_key(self, time: float) -> tuple[str, Hashable]:
        """Returns the state of the layer at the given time."""
        key = super().get_key(time)
        return (self.get_text(time), key)

    def _get_current_cursor_position(
        self, metrics: QFontMetrics, line: str, cursor_y: float,
        lineno: int, width: float
    ) -> tuple[float, float]:
        rect = metrics.boundingRect(line)
        if lineno == 0:
            cursor_y += rect.height()
        elif self.line_spacing is None:
            cursor_y += (rect.height() - rect.y())
        else:
            cursor_y += self.line_spacing

        if self.text_alignment == TextAlignment.LEFT:
            cursor_x = 0.
        elif self.text_alignment == TextAlignment.CENTER:
            cursor_x = (width - rect.width() - rect.x()) / 2
        elif self.text_alignment == TextAlignment.RIGHT:
            cursor_x = width - rect.width() - rect.x()
        else:
            raise ValueError(f"Invalid text alignment: {self.text_alignment}")
        return cursor_x, cursor_y

    def __call__(self, time: float) -> np.ndarray | None:
        if len(self.contents) == 0:
            return None
        text = self.get_text(time)
        if text is None or text == '':
            return None
        if not self._init_app:
            if QCoreApplication.instance() is None:
                QApplication(sys.argv[:1])
            self._init_app = True
        size = self.get_size(time)
        w, h = float(size[0]), float(size[1])

        # XXX: The returned rectangle and the area of the text
        # actually drawn seem to be different. We will keep
        # more margin in the interim
        margin = 40.
        max_stroke = _get_max_stroke(self.contents)
        W = np.floor(w + max_stroke + 2 * margin)
        H = np.floor(h + max_stroke + 2 * margin)
        image = QImage(W, H, QImage.Format.Format_ARGB32)
        max_color = _get_max_color(self.contents)
        if max_color is None:
            image.fill(QColor(0, 0, 0, 0))
        else:
            image.fill(QColor(*max_color, 0))

        painter = QPainter(image)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing, True)
        qfont = self._get_qfont(time)
        metrics = QFontMetrics(qfont)
        painter.setFont(qfont)
        lines = text.split('\n')
        for c in self.contents:
            if isinstance(c, FillProperty):
                r, g, b = c.color
                a = round(255 * c.opacity)
                painter.setPen(QColor(b, g, r, a))
                cursor_y = margin
                for i, line in enumerate(lines):
                    cursor_x, cursor_y = self._get_current_cursor_position(
                        metrics, line, cursor_y, lineno=i, width=w)
                    painter.drawText(QPointF(max_stroke + margin + cursor_x, cursor_y), line)
            elif isinstance(c, StrokeProperty):
                r, g, b = c.color
                a = round(255 * c.opacity)
                pen = QPen(
                    QColor(b, g, r, a), c.width, Qt.PenStyle.SolidLine,
                    Qt.PenCapStyle.RoundCap, Qt.PenJoinStyle.RoundJoin)
                painter.setPen(pen)
                painter_path = QPainterPath()
                cursor_y = margin
                for i, line in enumerate(lines):
                    cursor_x, cursor_y = self._get_current_cursor_position(
                        metrics, line, cursor_y, lineno=i, width=w)
                    painter_path.addText(QPointF(max_stroke + margin + cursor_x, cursor_y), qfont, line)
                painter.drawPath(painter_path)
        painter.end()
        array = qimage_to_numpy(image)
        return _clip_image(array)


def _clip_image(image: np.ndarray) -> np.ndarray:
    assert image.ndim == 3
    assert image.shape[2] == 4
    non_empty_pixels = np.all(image != np.array([0, 0, 0, 0]), axis=-1)
    non_empty_row_indices, non_empty_col_indices = np.where(non_empty_pixels)
    if non_empty_row_indices.size == 0 or non_empty_col_indices.size == 0:
        return image
    top, bottom = np.min(non_empty_row_indices), np.max(non_empty_row_indices)
    left, right = np.min(non_empty_col_indices), np.max(non_empty_col_indices)
    clipped_image = image[top: bottom + 1, left: right + 1]
    return clipped_image


def _get_max_stroke(contents: Sequence[FillProperty | StrokeProperty]) -> float:
    strokes = [c.width for c in contents if isinstance(c, StrokeProperty)]
    return float(max(strokes)) if 0 < len(strokes) else 0.


def _get_max_color(
    contents: Sequence[FillProperty | StrokeProperty]
) -> tuple[int, int, int] | None:
    strokes = [(c.width, c.color) for c in contents if isinstance(c, StrokeProperty)]
    if len(strokes) == 0:
        fills = [c.color for c in contents if isinstance(c, FillProperty)]
        return fills[-1] if 0 < len(fills) else None
    else:
        return max(strokes, key=lambda x: x[0])[1]



import cv2
import numpy as np

from ..attribute import Attribute, AttributesMixin, AttributeType
from ..enum import BlendingMode
from ..imgproc import alpha_composite


class GaussianBlur(AttributesMixin):
    """Gaussian blur effect.

    It applies Gaussian blur to the input image.

    .. note::
        The resolution of the output image is greater than
        that of the input image to preserve the blurring of the edges.

    Args:
        radius:
            Radius of Gaussian kernel.

    Animatable Attributes:
        ``radius``
    """

    def __init__(self, radius: float):
        self.radius = Attribute(radius, AttributeType.SCALAR, range=(0., 1e6))

    def __call__(self, prev_image: np.ndarray, time: float) -> np.ndarray:
        assert prev_image.ndim == 3
        assert prev_image.shape[2] == 4, f'prev_image must be RGBA image, but {prev_image.shape}'
        radius = float(self.radius(time))
        if radius == 0:
            return prev_image
        assert 0 < radius, f'radius must be nonnegative, but {radius}'

        k = 2 * int(np.ceil(max(1, (4 * radius - 1) / 2))) + 1
        ksize = (k, k)
        rgb_image = np.pad(prev_image[:, :, :3], (ksize, ksize, (0, 0)), mode='edge')
        alpha_image = np.pad(prev_image[:, :, 3:], (ksize, ksize, (0, 0)), mode='constant')
        return cv2.GaussianBlur(
            src=np.concatenate([rgb_image, alpha_image], axis=2),
            ksize=ksize, sigmaX=radius, sigmaY=radius)


class Glow(AttributesMixin):
    """Glow effect.

    It applies Gaussian blur to the input image and adds it to the original with the specified strength.

    Args:
        radius:
            Radius of Gaussian kernel.
        strength:
            Strength of the glow effect.

    Animatable Attributes:
        ``radius``
        ``strength``
    """

    def __init__(self, radius: float, strength: float = 1.0):
        self.radius = Attribute(radius, AttributeType.SCALAR, range=(0., 1e6))
        self.strength = Attribute(strength, AttributeType.SCALAR, range=(0., 100.))

    def __call__(self, prev_image: np.ndarray, time: float) -> np.ndarray:
        radius = float(self.radius(time))
        if radius == 0.:
            return prev_image

        assert 0 < radius, f'radius must be positive, but {radius}'
        k = 2 * int(np.ceil(max(1, (4 * radius - 1) / 2))) + 1
        ksize = (k, k)
        rgb_image = np.pad(prev_image[:, :, :3], (ksize, ksize, (0, 0)), mode='edge')
        alpha_image = np.pad(prev_image[:, :, 3:], (ksize, ksize, (0, 0)), mode='constant')
        pad_image = np.concatenate([rgb_image, alpha_image], axis=2)
        blurred_image = cv2.GaussianBlur(
            src=pad_image,
            ksize=ksize, sigmaX=radius, sigmaY=radius)
        strength = float(self.strength(time))
        bimg_rgb = blurred_image[:, :, :3].astype(np.float32)
        bimg_rgb = np.clip(strength * blurred_image[:, :, :3].astype(np.float32), 0, 255).astype(np.uint8)
        blurred_image = np.concatenate([bimg_rgb, blurred_image[:, :, 3:]], axis=2)
        return alpha_composite(
            pad_image, blurred_image, blending_mode=BlendingMode.LINEAR_DODGE)
